# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
from collections import Counter
from typing import Tuple, Type

import numpy as np
from mab import algs

from feijoa.search.oracles.meta.meta import MetaOracle
from feijoa.search.oracles.oracle import Oracle

log = logging.getLogger(__name__)


def relative_change_d1(x_cur: float, x_prev: float):
    """
    Return the relative difference between two points.

    Args:
        x_cur (float) :
            Current value.
        x_prev (float) :
            Previous value.

    Raises:
        AnyError: If anything bad happens.

    """

    sign = np.sign(x_prev - x_cur)
    return sign * np.abs(x_prev - x_cur) / (1e-10 + np.abs(x_cur) + np.abs(x_prev))


class UCB1(MetaOracle):
    """
    UCB1 bandit oracle.

    Used for the optimal choice of optimization strategy

    Args:
        algorithms:
            List of oracles.
        algo_select_count (int):
            Count of oracles to yield (with ranking).

    Raises:
        AnyError: If anything bad happens.

    """

    anchor = "UCB1"
    aliases: Tuple = ("UCB1", "ucb1", "ucb")

    STRATEGY: Type = algs.UCB1

    def __init__(self, *algorithms, algo_select_count=3):
        super().__init__(*algorithms)
        self.mab = None
        self.name2index = dict()
        self.results = []
        self.oracles_select_count: int = algo_select_count
        self.mult = 10
        self.rewards = Counter()

    def add_oracle(self, oracle):
        self.oracles.append(oracle)

    def remove_oracle(self, oracle: Oracle):
        self.oracles.remove(oracle)
        self._fix_names()
        self.mab = self.STRATEGY(len(self.oracles))
        self.name2index = {oracle.name: i for i, oracle in enumerate(self.oracles)}

        for algo_index, reward in self.rewards.items():
            for _ in range(reward):
                self.mab.reward(algo_index)

    def _fix_names(self):
        """Needed for make unique oracles names."""

        names = set()
        for a in self.oracles:
            while a.name in names:
                a.name += "@"
            names.add(a.name)

    @property
    def order(self):
        """
        Get order for oracles.

                Raises:
                    AnyError: If anything bad happens.

        """

        if not self.mab:
            self._fix_names()
            self.mab = self.STRATEGY(len(self.oracles))
            self.name2index = {oracle.name: i for i, oracle in enumerate(self.oracles)}

        _, ranking = self.mab.select()
        index = max(len(self.oracles), self.oracles_select_count)
        ranking_chunk = ranking[:index]

        return [self.oracles[i] for i in ranking_chunk]

    def tell(self, config, result):
        """
        Tell results to all search oracles.

                An adaptive strategy is used with increasing
                reward upon convergence.

                Raises:
                    AnyError: If anything bad happens.

        """

        reward_multiplier = 1

        if len(self.results) % 10 == 0:
            log.debug(f"Up reward multiplier to {self.mult}")
            self.mult = int(self.mult * 1.2)

        has_reward = False
        oracle_index = None

        if np.isfinite(result) and config.requestor in self.name2index:
            oracle_index = self.name2index[config.requestor]

            if self.results:
                m = min(self.results)

                relative_change = relative_change_d1(result, m)
                has_reward = relative_change > 0.005
                performance = relative_change * self.mult

                if np.isfinite(performance):
                    reward_multiplier = max(1, int(relative_change * self.mult))
                else:
                    reward_multiplier = 100

        if has_reward:
            for _ in range(reward_multiplier):
                self.mab.reward(oracle_index)
                self.rewards[oracle_index] += 1
            log.debug(
                f"Reward <{reward_multiplier}> "
                f"for {config.requestor} with result: {result}"
            )

        self.results.append(result)

        for oracle in self.oracles:
            oracle.tell(config, result)


class UCBTuned(UCB1):

    anchor = "UCBTuned"
    aliases = ("UCBTuned", "ucbtuned")

    STRATEGY = algs.UCBTuned


class ThompsonSampler(UCB1):

    anchor = "ThompsonSampler"
    aliases = ("ThompsonSampler", "thompson", "th")

    STRATEGY = algs.ThompsomSampling



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Evaluations plot module."""

import plotly.graph_objs as go
from plotly.subplots import make_subplots


def pad(string: str, width: int, filler=" ", fill_chars=3):
    """
    Pads a string to a given width.

    Args:
        string (str):
            String instance.
        width (int):
            Target width of string.
        filler:
            Fill places with specified char.
        fill_chars:
            Count of chars to fill.

    Raises:
        AnyError: If anything bad happens.

    """

    if len(string) <= width:
        return string
    return f"{string[:width - 3]}{filler * fill_chars}"


def plot_evaluations(job, params=None, invert_objective=False):
    """
    Plot evaluations of job.

    Args:
        job (Job):
            Job instance.
        params (list):
            List of parameters to plot.
        invert_objective (bool):
            For the maximization task, it is
            necessary to invert the value of
            the view function, in this case,
            you must check this box to get
            corrective values

    Raises:
        AnyError: If anything bad happens.

    """

    df = job.get_dataframe(brief=True, only_good=True)
    objectives = df["objective_result"] * (1 if not invert_objective else -1)
    df.drop(columns=["id", "objective_result"], inplace=True)
    params = params or df.columns

    if any([param not in df.columns for param in params]):
        raise ValueError("Unknown columns in specified job.")

    params_length = len(params)

    fig = make_subplots(rows=params_length, cols=params_length)

    for i, col_x in enumerate(params):
        for j, col_y in enumerate(params):
            if i > j:
                continue

            if i == j:
                fig.add_histogram(
                    x=df[col_x],
                    nbinsx=30,
                    row=i + 1,
                    col=i + 1,
                    marker=dict(
                        color="#330C73",
                    ),
                    opacity=0.75,
                )
                continue

            x = df[col_x]
            y = df[col_y]

            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=y,
                    mode="markers",
                    marker=dict(
                        color=objectives,
                        colorscale="TealRose_r",
                        symbol="circle",
                        colorbar=dict(title="Objective result"),
                        line=dict(width=1, color="black"),
                    ),
                ),
                col=i + 1,
                row=j + 1,
            )

    fig.update_layout(
        title="Evaluations plot",
        showlegend=False,
        bargap=0.2,
        bargroupgap=0.1,
    )

    def cut(base, num):
        if num == 1:
            return base
        return base if num == 1 else f"{base}{num}"

    positions = [
        (
            *fig["layout"][cut("xaxis", i)]["domain"],
            *fig["layout"][cut("yaxis", i)]["domain"],
        )
        for i in range(1, params_length**2 + 1)
    ]

    positions = [
        (x0, x1, y0, y1) for (x0, x1, y0, y1) in positions if x0 == 0.0 or y0 == 0.0
    ]

    names = list(params) * 2

    for i, (x0, x1, y0, y1) in enumerate(positions):
        x = x0
        y = y0
        angle = 0

        if x == 0.0:
            x -= 0.05
            y = (y0 + y1) / 2
            angle = 270
        else:
            x = (x0 + x1) / 2 + 0.02
            y -= 0.08

        fig.add_annotation(
            dict(
                x=x,
                y=y,
                xref="paper",
                yref="paper",
                text=pad(names.pop(0), 10, filler="."),
                showarrow=False,
            ),
            textangle=angle,
        )

        if x0 == 0 and y0 == 0:
            x_ = (x0 + x1) / 2
            y_ = -0.08
            angle = 0
            fig.add_annotation(
                dict(
                    x=x_,
                    y=y_,
                    xref="paper",
                    yref="paper",
                    text=pad(names.pop(0), 10, filler="."),
                    showarrow=False,
                ),
                textangle=angle,
            )

    return fig



import sys


def in_notebook():
    """
    Check if module in running in IPython kernel.

        Returns:
             ``True`` if the module is running in IPython kernel,
             ``False`` if in IPython shell or other Python shell.

        Raises:
            AnyError: If anything bad happens.

    """
    return "ipykernel" in sys.modules



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import importlib
import types
from typing import Any

from feijoa.exceptions import PackageNotInstalledError


class ImportWrapper:
    """Context manager to wrap import messages."""

    def __enter__(self):
        pass

    def __exit__(self, exctype, excinst, exctb):
        if exctype == ModuleNotFoundError:
            module = str(excinst).split()[-1]
            message = (
                f"Tried to import {module} but failed."
                " Please make sure that the package is"
                " installed correctly to use this feature."
            )
            raise PackageNotInstalledError(message)
        return True


def import_or_skip(module, reason=""):
    """
    Import module or skip test in
        pytest framework.

        Args:
            module:
                Python module.
            reason:
                Reason of skipping.

        Raises:
            AnyError: If anything bad happens.

    """
    import pytest

    try:
        return pytest.importorskip(module, reason=reason)
    except PackageNotInstalledError:
        return pytest.skip(reason=reason, allow_module_level=True)


class LazyModuleImportProxy(types.ModuleType):
    """Class for lazy modules importing."""

    def __init__(self, module: str) -> None:
        super().__init__(module)
        self.module = module

    def _load(self) -> types.ModuleType:
        module = importlib.import_module(self.module)
        self.__dict__.update(module.__dict__)
        return module

    def __getattr__(self, name: str) -> Any:
        return getattr(self._load(), name)



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""SQLAlchemy models module."""

import json

from sqlalchemy import Column, Float, ForeignKey, Integer, String, TypeDecorator, types
from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base
from sqlalchemy.orm import relationship

__all__ = [
    "SearchSpaceModel",
    "ParameterModel",
    "JobModel",
    "ExperimentModel",
]


_Base: DeclarativeMeta = declarative_base()


class Json(TypeDecorator):  # type: ignore
    """Json column (temporary class)."""

    @property
    def python_type(self):
        return object

    impl = types.String

    def process_bind_param(self, value, dialect):
        return json.dumps(value)

    def process_literal_param(self, value, dialect):
        return value

    def process_result_value(self, value, dialect):
        try:
            return json.loads(value)
        except (ValueError, TypeError):
            return None


class SearchSpaceModel(_Base):  # type: ignore
    __tablename__ = "search_space"

    id = Column(Integer, primary_key=True)


class ParameterModel(_Base):  # type: ignore
    __tablename__ = "parameter"

    id = Column(Integer, primary_key=True)
    search_space_id = Column(ForeignKey(SearchSpaceModel.id))  # type: ignore
    search_space = relationship(SearchSpaceModel, backref="parameters")  # type: ignore

    name = Column(String)
    kind = Column(String)
    meta = Column(Json)


class JobModel(_Base):  # type: ignore
    __tablename__ = "jobs"

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)
    search_space_id = Column(ForeignKey(SearchSpaceModel.id))  # type: ignore
    search_space = relationship(SearchSpaceModel, backref="jobs")  # type: ignore
    last_optimizer = Column(String)

    def __repr__(self):
        return f"Job<{self.id}, {self.name}"


class ExperimentModel(_Base):  # type: ignore
    __tablename__ = "experiments"

    id = Column(Integer, primary_key=True)
    job_id = Column(ForeignKey(JobModel.id), primary_key=True)  # type: ignore
    job = relationship(JobModel, backref="experiments")  # type: ignore
    state = Column(String)
    hash = Column(String)
    objective_result = Column(Float)
    requestor = Column(String)
    params = Column(Json)
    create_timestamp = Column(Float)
    finish_timestamp = Column(Float)
    metrics = Column(Json)



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Parallel coordinates plot module."""

import numpy as np
import plotly.graph_objs as go
from sklearn.preprocessing import LabelEncoder


def pad(string: str, width: int, filler=" ", fill_chars=3):
    """
    Pads a string to a given width.

    Args:
        string (str):
            String instance.
        width (int):
            Target width of string.
        filler:
            Fill places with specified char.
        fill_chars:
            Count of chars to fill.

    Raises:
        AnyError: If anything bad happens.

    """

    if len(string) <= width:
        return string
    return f"{string[:width - 3]}{filler * fill_chars}"


def plot_parallel_coordinates(
    job,
    params=None,
    invert_objective=False,
):
    """
    Plot the parallel coordinates for
    a specified job.

    Args:
        job (Job):
            Job instance.
        params (list):
            Parameters names to plot.
        invert_objective (bool):
            For the maximization task, it is
            necessary to invert the value of
            the view function, in this case,
            you must check this box to get
            corrective values

    Raises:
        AnyError: If anything bad happens.

    """

    df = job.get_dataframe(brief=True, only_good=True)
    objectives = df["objective_result"] * (1 if not invert_objective else -1)
    df.drop(columns=["id", "objective_result"], inplace=True)
    omin = objectives.min()
    omax = objectives.max()
    params = params or df.columns

    if any([param not in df.columns for param in params]):
        raise ValueError("Unknown columns in specified job.")

    dimensions = [
        {
            "label": "objective_result",
            "values": objectives,
            "range": (omin, omax),
        }
    ]

    if any([df[col].dtype.name == "object" for col in params]):
        raise ValueError("Dataframe should not have columns with object type.")

    for col in sorted(params, reverse=True):
        dim = {
            "label": pad(col, 15, "."),
        }

        if df[col].dtype.name == "category":
            ticks = LabelEncoder().fit_transform(df[col])
            ticks_text = df[col]

            dim["values"] = ticks
            dim["tickvals"] = np.arange(len(set(ticks)))
            dim["ticktext"] = ticks_text
            dim["range"] = ticks.min(), ticks.max()
        else:
            dim["values"] = df[col]
            dim["range"] = df[col].min(), df[col].max()

        dimensions.append(dim)

    tr = [
        go.Parcoords(
            dimensions=dimensions,
            line=dict(
                color=objectives,
                colorscale="TealRose",
                colorbar=dict(title="objective_result"),
                showscale=True,
            ),
        )
    ]

    layout = go.Layout(title="Parallel Coordinate Plot")
    fig = go.Figure(data=tr, layout=layout)
    return fig



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Round robin meta oracle module."""

import logging
from collections import deque

from feijoa.search.oracles.meta.meta import MetaOracle
from feijoa.search.oracles.oracle import Oracle

log = logging.getLogger(__name__)

__all__ = ["RoundRobinMeta"]


class RoundRobinMeta(MetaOracle):
    """
    RoundRobin meta oracle.

    Raises:
        AnyError: If anything bad happens.

    """

    anchor = "roundrobin"
    aliases = ("roundrobin",)

    def __init__(self, *algorithms):
        super().__init__(*algorithms)
        self.oracle_deque = None

    @property
    def order(self):
        # noinspection PyUnresolvedReferences
        if not self.oracle_deque:
            self.oracle_deque = deque(self.oracles)

        self.oracle_deque.rotate(1)
        return self.oracle_deque

    def remove_oracle(self, oracle: Oracle):
        self.oracle_deque.remove(oracle)
        self.oracles.remove(oracle)



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Different mixins module."""
from abc import ABCMeta, abstractmethod


class Observer(metaclass=ABCMeta):
    """
    The Observer interface declares the
    update method, used by subjects.
    """

    @abstractmethod
    def update(self, event, subject, *args, **kwargs):
        """
        Receive update from subject.

                Args:
                    event:
                        Type of event.
                    subject:
                        Subject of notification.

                Returns:
                    None

                Raises:
                    AnyError: If anything bad happens.

        """

        pass


class Subject(metaclass=ABCMeta):
    """
    The Subject interface declares a set
    of methods for managing subscribers.
    """

    @abstractmethod
    def attach(self, observer: Observer) -> None:
        """
        Attach an observer to the subject.
        """
        pass

    @abstractmethod
    def detach(self, observer: Observer) -> None:
        """
        Detach an observer from the subject.
        """
        pass

    @abstractmethod
    def notify(self, event, *args, **kwargs) -> None:
        """
        Notify all observers about an event.
        """
        pass



import io
from contextlib import redirect_stderr

from ply import lex, yacc

__all__ = ["_OracleParser"]


def flatten(x):
    if isinstance(x, (list, tuple, set)):
        return [a for i in x for a in flatten(i)]
    else:
        return [x]


# noinspection SpellCheckingInspection,PyPep8Naming,PyMethodMayBeStatic
class _OracleLexer:
    tokens = (
        "WORD",
        "ANGLEBRACKETL",
        "ANGLEBRACKETR",
        "SQBRACKETL",
        "SQBRACKETR",
        "COMMA",
        "PLUS",
        "EQ",
        "INTEGER",
        "FLOAT",
    )

    ident = r"[a-zA-Z_]\w*"

    t_WORD = ident
    t_ANGLEBRACKETL = r"\<"
    t_ANGLEBRACKETR = r"\>"
    t_SQBRACKETL = r"\["
    t_SQBRACKETR = r"\]"
    t_PLUS = r"\+"
    t_COMMA = r"\,"
    t_EQ = r"\="

    def t_FLOAT(self, t):
        r"""[-+]?[0-9]+(\.([0-9]+)?([eE][-+]?[0-9]+)?|[eE][-+]?[0-9]+)"""
        t.value = float(t.value)
        return t

    def t_INTEGER(self, t):
        r"""\d+"""
        t.value = int(t.value)
        return t

    t_ignore = " \r\n\t\f"

    def t_error(self, t):
        self.exception_message = (
            f"Illegal character `{t.value[0]}` on {t.lineno} line,"
            f" {t.lexpos} position"
        )
        t.lexer.skip(1)

    def t_newline(self, t):
        r"""\n+"""
        t.lexer.lineno += len(t.value)

    def __init__(self):
        self.lexer = lex.lex(module=self)
        self.exception_message = ""


# noinspection PyMethodMayBeStatic,DuplicatedCode
class _OracleParser:
    tokens = _OracleLexer.tokens

    def p_super_oracle(self, p):
        """
        super_oracle : WORD ANGLEBRACKETL oracle-list ANGLEBRACKETR
                     | WORD ANGLEBRACKETL oracle-list ANGLEBRACKETR SQBRACKETL params-list SQBRACKETR
                     | oracle
        """

        if len(p) == 2:
            p[0] = {"top-oracle": p[1]["name"], **p[1]}
            p[0].pop("name")
        if len(p) == 5:
            p[0] = {"top-oracle": p[1], "oracles": p[3]}
        elif len(p) == 8:
            p[0] = {
                "top-oracle": p[1],
                "oracles": p[3],
                "params": p[6],
            }

    def p_oracle(self, p):
        """
        oracle : WORD
               | WORD SQBRACKETL params-list SQBRACKETR
               | WORD PLUS plugins-list
               | WORD SQBRACKETL params-list SQBRACKETR PLUS plugins-list
        """

        if len(p) == 2:
            p[0] = {"name": p[1]}
        elif len(p) == 4:
            p[0] = {"name": p[1], "plugins": p[3]}
        elif len(p) == 5:
            p[0] = {"name": p[1], "params": p[3]}
        elif len(p) == 7:
            p[0] = {"name": p[1], "params": p[3], "plugins": p[6]}

    def p_plugin(self, p):
        """
        plugin : WORD
               | WORD SQBRACKETL params-list SQBRACKETR
        """

        if len(p) == 2:
            p[0] = {"name": p[1]}
        elif len(p) == 5:
            p[0] = {"name": p[1], "params": p[3]}

    def p_plugins_list(self, p):
        """
        plugins-list : plugin
                     | plugins-list PLUS plugin
        """

        p[0] = [p[i] for i in range(1, len(p), 2)]
        p[0] = flatten(p[0])

    def p_oracle_list(self, p):
        """
        oracle-list : oracle
                    | oracle-list COMMA oracle
        """

        params = [p[i] for i in range(1, len(p), 2)]
        total_oracles = []

        def iterate(obj):
            nonlocal total_oracles

            if isinstance(obj, dict):
                total_oracles.append(obj)

            if isinstance(obj, list):
                for o in obj:
                    iterate(o)

        iterate(params)

        p[0] = total_oracles

    def p_value(self, p):
        """
        value : INTEGER
              | FLOAT
              | WORD
        """

        p[0] = p[1]

    def p_param(self, p):
        """param : WORD EQ value"""

        p[0] = {p[1]: p[3]}

    def p_params_list(self, p):
        """
        params-list : param
                    | params-list COMMA param
        """

        params = [p[i] for i in range(1, len(p), 2)]

        total_params = {}

        def iterate(obj):
            nonlocal total_params

            if isinstance(obj, dict):
                if "params" in obj.keys():
                    for o in obj["params"]:
                        iterate(o)
                else:
                    total_params = {**total_params, **obj}

            if isinstance(obj, list):
                for o in obj:
                    iterate(o)

        iterate(params)

        p[0] = total_params

    def p_error(self, p):
        if p is None:
            return
        self.exception_message = (
            f"Error parsing `{p.value}` " f"at {p.lineno} line in {p.lexpos} position."
        )

    def __init__(self):

        self.exception_message = ""
        self.lexer = _OracleLexer()

        # suppress ply logs
        with io.StringIO() as buf, redirect_stderr(buf):
            self.parser = yacc.yacc(module=self)

    def parse(self, doc):
        self.exception_message = ""

        with io.StringIO() as buf, redirect_stderr(buf):
            tree = self.parser.parse(doc, lexer=self.lexer.lexer, debug=1)

            if self.lexer.exception_message:
                # intercept lexer error message
                self.exception_message = self.lexer.exception_message

            # try pick up error message from lexer debug output
            if not tree and not self.exception_message:
                error_line = next(
                    (line for line in buf.getvalue().splitlines() if "ERROR" in line),
                )
                self.exception_message = (
                    "Error while parsing at end of line. Description:\n"
                    f"\t{error_line}"
                )

        if self.exception_message:
            raise SyntaxError(self.exception_message)

        return tree



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Base importance evaluator class module."""

import abc

from feijoa.jobs.job import Job

__all__ = ["ImportanceEvaluator"]


class ImportanceEvaluator(metaclass=abc.ABCMeta):
    """
    Base importance evaluator.

    Measures the excessive amount of hyperparameter
    taken per feature value. Helps to select only the most
    influential hyperparameters for their tuning.
    Does not require additional measurements.

    .. note::
        This version does not apply the study of the influence
        of a combination of hyperparameters.
    """

    @abc.abstractmethod
    def do(self, job: Job):
        """
        Calculate hyperparameters importance.

        Args:
            job (Job) : Specified job.

        Returns:
            Dict with keys `params` and `importances`
                `params` contains names of parameters,
                `importances` contains list of importances.

        Raises:
            AnyError: If anything bad happens.
        """

        raise NotImplementedError()



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Optimization history plot module."""
import numpy as np
import plotly.graph_objs as go


def plot_optimization_history(
    job,
    mode="lines+markers",
    name=None,
    fig=None,
    only_best=True,
    invert_objective=False,
):
    """
    Plot optimization history.

    Args:
        job (Job):
            Job instance.
        mode (str):
            Plotly plot mode. Defaults to "lines+markers"
        name (str):
            Name of trace.
        fig (go.Figure):
            Plotly figure object.
        only_best:
            Plotly only bests results.
            (Descending sequence by objective values)
        invert_objective (bool):
            For the maximization task, it is
            necessary to invert the value of
            the view function, in this case,
            you must check this box to get
            corrective values


    Raises:
        AnyError: If anything bad happens.

    """

    df = job.get_dataframe()
    obj = df["objective_result"] * (1 if not invert_objective else -1)
    iterations = df["id"]
    name = name or job.name

    convergence = np.zeros_like(df["objective_result"])

    m = float("+inf")
    for i in range(len(convergence)):
        if obj[i] < m:
            m = obj[i]
        convergence[i] = m

    fig = fig or go.Figure()

    fig.add_trace(
        go.Scatter(
            x=iterations,
            y=convergence,
            mode=mode,
            name=f"{name} bests",
        )
    )

    if not only_best:
        full_df = job.get_dataframe()
        fig.add_trace(
            go.Scatter(
                x=full_df["id"],
                y=full_df["objective_result"],
                mode="markers",
                name=name,
            )
        )

    fig.update_layout(
        title="Optimization history plot",
        xaxis_title="Iteration",
        yaxis_title="Objective value",
    )

    return fig


def plot_compare_jobs(
    *jobs,
    mode="lines+markers",
    fig=None,
    names=None,
    invert_objective=False,
):
    """
    Plot comparison between jobs.

    Args:
        jobs (List[Job]):
            List of jobs.
        mode (str):
            Plotly plot mode. Defaults to "lines+markers"
        names (List[str]):
            Names of traces.
        fig (go.Figure):
            Plotly figure object.
        invert_objective (bool):
            For the maximization task, it is
            necessary to invert the value of
            the view function, in this case,
            you must check this box to get
            corrective values

    Raises:
        AnyError: If anything bad happens.

    """

    if isinstance(names, list) and len(names) != len(jobs):
        raise ValueError(f"Names must contains {len(jobs)} values.")

    names = names or [job.name for job in jobs]
    print(names)

    fig = fig or go.Figure()

    for job, name in zip(jobs, names):
        plot_optimization_history(job, mode, name, fig)

    return fig



import contextlib
import importlib.util
import inspect
import pathlib
import sys
from functools import lru_cache
from itertools import chain

from feijoa import __feijoa_folder__
from feijoa.exceptions import PackageNotInstalledError, SearchOracleNotFoundedError
from feijoa.plugins.plugin import Plugin
from feijoa.search.oracles.meta.bandit import UCB1
from feijoa.search.oracles.meta.meta import MetaOracle
from feijoa.search.oracles.oracle import Oracle
from feijoa.search.space import SearchSpace

# noinspection PyProtectedMember
from feijoa.utils._oracle_parser import _OracleParser

SEARCH_FOLDER = pathlib.Path(__feijoa_folder__) / "search"
ALGORITHMS_FOLDER = SEARCH_FOLDER / "oracles"
PLUGINS_FOLDER = pathlib.Path(__feijoa_folder__) / "plugins"
INTEGRATION_FOLDER = pathlib.Path(__feijoa_folder__).parent / "integration"


@lru_cache(maxsize=None)
def fetch_classes(base_class, *folders, only_anchors=False):
    """
    Fetch oracle from specified oracles folders.

    By default, uses:
        - default oracles folder

    :returns: all founded oracle's classes.
    """

    assert len(folders) >= 1, "Folders must be passed"

    pool = dict()

    target = (
        script
        for script in chain(*[pathlib.Path(folder).rglob("*.py") for folder in folders])
        if "__" not in script.name
    )

    for script in target:
        spec = importlib.util.spec_from_file_location(script.stem, str(script))
        module = importlib.util.module_from_spec(spec)
        sys.modules[module.__name__] = module

        with contextlib.suppress(PackageNotInstalledError):
            spec.loader.exec_module(module)

        iterable = (
            obj
            for obj in inspect.getmembers(module, inspect.isclass)
            if (obj[1].__module__ == module.__name__ and issubclass(obj[1], base_class))
        )

        for name, cls in iterable:
            pool[cls.anchor] = cls

            if not only_anchors:
                pool[name] = cls

                with contextlib.suppress(TypeError):
                    for al in cls.aliases:
                        pool[al] = cls

    return pool


def fetch_oracles(include_integration=False):
    args = [ALGORITHMS_FOLDER]
    if include_integration:
        args.append(INTEGRATION_FOLDER)

    oracles = fetch_classes(Oracle, *args)
    return oracles


def fetch_plugins(include_integration=False):
    args = [ALGORITHMS_FOLDER]
    if include_integration:
        args.append(INTEGRATION_FOLDER)
    plugins = fetch_classes(Plugin, *args)
    return plugins


def fetch_top_oracles(include_integration=False):
    args = [SEARCH_FOLDER]
    if include_integration:
        args.append(INTEGRATION_FOLDER)
    top_oracles = fetch_classes(MetaOracle, *args)
    return top_oracles


def get_algo(name):
    """Get oracle class by name"""

    classes = fetch_oracles()
    cls = classes.get(name)

    if not cls:
        raise SearchOracleNotFoundedError()

    return cls


def get_top_oracle(name):

    classes = fetch_top_oracles()
    cls = classes.get(name)

    if not cls:
        raise SearchOracleNotFoundedError()

    return cls


def get_plugin(name):
    classes = fetch_plugins()
    cls = classes.get(name)

    if not cls:
        raise SearchOracleNotFoundedError()

    return cls


def maker(line, search_space: SearchSpace, random_state=None):
    parser = _OracleParser()
    parsed = parser.parse(line)

    top_oracle = parsed["top-oracle"]
    top_oracle_params = parsed.get("params", {})
    top_oracle_plugins = parsed.get("plugins", [])
    oracles = parsed.get("oracles")

    if not oracles:
        top_oracle_cls = get_algo(top_oracle)
        # if passed only oracle

        plugins = []

        for plug in top_oracle_plugins:
            plug_name = plug["name"]
            plug_params = plug.get("params", {})
            plug_cls = get_plugin(plug_name)
            plugins.append(plug_cls(**plug_params))

        oracle_instance = top_oracle_cls(
            search_space=search_space,
            **top_oracle_params,
            seed=random_state,
        )

        for p in plugins:
            p.subscribers = [oracle_instance]

        top_oracle_instance = UCB1(oracle_instance)

        return top_oracle_instance

    top_oracle_cls = get_top_oracle(top_oracle)

    top_oracle_params = parsed.get("params", {})

    oracles_pool = []

    for oracle in oracles:
        o_name = oracle["name"]
        o_plugins = oracle.get("plugins", [])
        o_params = oracle.get("params", {})
        o_cls = get_algo(o_name)

        plugins = []

        for plug in o_plugins:
            plug_name = plug["name"]
            plug_params = plug.get("params", {})
            plug_cls = get_plugin(plug_name)
            plugins.append(plug_cls(**plug_params))

        oracle_instance = o_cls(
            search_space=search_space, **o_params, seed=random_state
        )

        for p in plugins:
            p.subscribers = [oracle_instance]

        oracle_instance.subscribers = plugins

        oracles_pool.append(oracle_instance)

    top_oracle_instance = top_oracle_cls(*oracles_pool, **top_oracle_params)
    return top_oracle_instance



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Genetic oracles with `pymoo` backend module."""

import logging
import time
from typing import Generator, List, Optional, Tuple

import numpy as np

from feijoa.utils.imports import ImportWrapper, LazyModuleImportProxy

with ImportWrapper():
    from pymoo.core.problem import Problem
    from pymoo.core.termination import NoTermination
    from pymoo.core.individual import Individual
    from pymoo.core.population import Population
    from pymoo.algorithms.base.local import LocalSearch

from feijoa.models.configuration import Configuration
from feijoa.search.oracles.oracle import Oracle
from feijoa.utils.transformers import inverse_transform, transform

de = LazyModuleImportProxy("pymoo.algorithms.soo.nonconvex.de")
ga = LazyModuleImportProxy("pymoo.algorithms.soo.nonconvex.ga")
cmaes = LazyModuleImportProxy("pymoo.algorithms.soo.nonconvex.cmaes" "")
brkga = LazyModuleImportProxy("pymoo.algorithms.soo.nonconvex.brkga")

direct = LazyModuleImportProxy("pymoo.algorithms.soo.nonconvex.direct")
es = LazyModuleImportProxy("pymoo.algorithms.soo.nonconvex.es")
nichega = LazyModuleImportProxy("pymoo.algorithms.soo.nonconvex.ga_niching")
g3pcs = LazyModuleImportProxy("pymoo.algorithms.soo.nonconvex.g3pcx")
isres = LazyModuleImportProxy("pymoo.algorithms.soo.nonconvex.isres")
pso = LazyModuleImportProxy("pymoo.algorithms.soo.nonconvex.pso")
eppso = LazyModuleImportProxy("pymoo.algorithms.soo.nonconvex.pso_ep")
sres = LazyModuleImportProxy("pymoo.algorithms.soo.nonconvex.sres")


log = logging.getLogger(__name__)


class Genetic(Oracle):

    """
    Genetic oracles with `pymoo` backend.
    Pymoo link: https://pymoo.org/.

    Args:
        algorithm_cls:
            Pymoo oracle class.

    Raises:
        AnyError: If anything bad happens.

    .. note::
        This set of oracles is experimental.

    """

    anchor: str = "GA"
    aliases: Tuple = ("Genetic", "genetic", "GA", "ga")

    algorithm_cls = ga.GA

    def __init__(self, search_space, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.search_space = search_space

        # parse lower and upper bounds
        self.xl = self.search_space.bounds[:, 0]
        self.xu = self.search_space.bounds[:, 1]

        self._ask_gen = None

        self.problem = Problem(
            n_var=len(self.search_space),
            n_obj=1,
            n_constr=0,
            xl=self.xl,
            xu=self.xu,
        )

        self.algorithm = self.algorithm_cls(**kwargs)

        # let the oracle object never terminate and let the loop control it
        self.termination = NoTermination()

        # create an oracle object that never terminates
        self.algorithm.setup(self.problem, termination=self.termination)

        self.args = args
        self.kwargs = kwargs

        self.algorithm_cls.start_time = time.time()

        # fix the random seed manually
        np.random.seed(0)

        self.pool = []

        self.pop = None

    def ask(self, n: int = 1) -> Optional[List[Configuration]]:
        if not self._ask_gen:
            self._ask_gen = self._ask(n)
            log.debug(
                "Parameter `n` does not affect on configuration's count,"
                f" because {self.__class__.__name__} is can have variable pop size."
            )
        return next(self._ask_gen)

    # noinspection PyUnusedLocal
    def _ask(self, n: int) -> Generator:
        """Main ask generator for genetic oracles."""

        n_gen = 0

        while True:
            self.pop = self.algorithm.ask()

            # get the design space values of the oracle
            X = self.pop.get("X")

            log.debug(f"Generation: {n_gen} for {self.name}")

            configs = [
                Configuration(
                    transform(
                        solution,
                        self.search_space,
                    ),
                    requestor=self.name,
                    request_id=i,
                )
                for i, solution in enumerate(X)
            ]

            yield configs

            self.algorithm.tell(infills=self.pop)

            n_gen += 1

    def tell(self, config, result):

        # take our results

        if config.requestor == self.name:
            self.pop[config.request_id].F = np.array([result])
            self.pool.append(self.pop[config.request_id])
            return

        # take experience from other oracles

        x0 = Individual()

        # cook our individual instance
        x0.X = inverse_transform(config, self.search_space)
        x0.F = np.array([result])

        if self.pool and result < min(self.pool, key=lambda solution: solution.F).F:
            log.debug(f"Restarting {self.name}")

            # take into account base class type
            params = {"sampling": Population(individuals=np.array(self.pool))}
            if issubclass(self.algorithm_cls, LocalSearch):
                params = {"x0": Population(individuals=np.array(self.pool))}

            # restart our oracle
            self.algorithm = self.algorithm_cls(*self.args, **self.kwargs, **params)
            self.algorithm.setup(self.problem, termination=self.termination)

        self.pool.append(x0)


class DifferentialEvolution(Genetic):
    @property
    def algorithm_cls(self):
        return de.DE

    anchor = "de"
    aliases = ("differentialevolution", "DE")


class CMAES(Genetic):
    @property
    def algorithm_cls(self):
        return cmaes.CMAES

    anchor = "cmaes"
    aliases = ("cmaes",)


class BRKGA(Genetic):
    @property
    def algorithm_cls(self):
        return brkga.BRKGA

    anchor = "brkga"
    aliases = ("brkga",)


class DIRECT(Genetic):
    @property
    def algorithm_cls(self):
        return direct.DIRECT

    anchor = "direct"
    aliases = ("direct",)


class ES(Genetic):
    @property
    def algorithm_cls(self):
        return es.ES

    anchor = "es"
    aliases = ("es",)


class G3PCX(Genetic):
    @property
    def algorithm_cls(self):
        return g3pcs.G3PCX

    anchor = "g3pcx"
    aliases = ("g3pcx",)


class NICHEGA(Genetic):
    @property
    def algorithm_cls(self):
        return nichega.NicheGA

    anchor = "nichega"
    aliases = ("niche", "nichega")


class PSO(Genetic):
    @property
    def algorithm_cls(self):
        return pso.PSO

    anchor = "pso"
    aliases = ("pso",)


class EPPSO(Genetic):
    @property
    def algorithm_cls(self):
        return eppso.EPPSO

    anchor = "eppso"
    aliases = ("eppso",)


class SRES(Genetic):
    @property
    def algorithm_cls(self):
        return sres.SRES

    anchor = "sres"
    aliases = ("sres",)


class ISRES(Genetic):
    @property
    def algorithm_cls(self):
        return isres.ISRES

    anchor = "isres"
    aliases = ("isres",)



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Parameters module."""

import abc

__all__ = [
    "Parameter",
    "Integer",
    "Real",
    "Categorical",
    "ParametersVisitor",
]

import numpy as np

from feijoa.exceptions import ParametersIncorrectInputValues


class Parameter(metaclass=abc.ABCMeta):
    """Base class of parameter."""

    def __init__(self, name: str):
        self.name = name

    def __repr__(self):
        raise NotImplementedError()

    def __str__(self):
        return repr(self)

    @abc.abstractmethod
    def accept(self, v):
        """
        Accept visitor.

        Args:
            v (Visitor):
                ParametersVisitor instance.

        Returns:
            Any

        """

        raise NotImplementedError()

    @abc.abstractmethod
    def seed(self):
        """Make some legal value."""

        raise NotImplementedError()

    @abc.abstractmethod
    def is_primitive(self):
        """Primitive parameter, such as integer, real."""

        raise NotImplementedError()

    @abc.abstractmethod
    def get_unit_value(self, value):
        """Only for primitive parameters."""

        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def meta(self):
        """Meta information of parameter."""

        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def bounds(self):
        """Bounds of parameter."""

        raise NotImplementedError()


class Integer(Parameter):
    """
    Integer parameter class.

    Represents discrete integer parameters from Z-space.

    Such parameters can be, for example, the number of
    layers in neural networks or the number of neurons
    in a hidden layer, max depth of tree.

    Example:

    .. code-block:: python

        from feijoa.search.parameters import Integer

        param = Integer("foo", low=0, high=10)

    .. note:
        In parameters creation strict-typing is used
        for input values.

    Args:
        name (str):
            Name of parameter.
        low (int):
            Lower bound of parameter.
        high (int):
            Upper bound of parameter.

    Raises:
        ParametersIncorrectInputValues:
            If input values is not valid integer values.
        AnyError:
            If anything bad happens.

    """

    def __init__(self, name: str, *, low: int, high: int):
        super().__init__(name)

        if not isinstance(low, int) or not isinstance(high, int):
            raise ParametersIncorrectInputValues(
                "Low and High bounds for" " Integer parameter" " must be ints."
            )
        if low > high:
            raise ParametersIncorrectInputValues(
                "High bound must" " be greater than low bound."
            )

        self.low = low
        self.high = high

    def __repr__(self):
        return f"Integer('{self.name}, low={self.low}, high={self.high})"

    @property
    def bounds(self):
        return np.array([self.low, self.high])

    def accept(self, v):
        return v.visit_integer(self)

    def seed(self):
        return self.low

    def is_primitive(self):
        return True

    def get_unit_value(self, value):
        return float(value - (self.low - 0.4999)) / float(
            (self.high + 0.4999) - (self.low - 0.4999)
        )

    @property
    def meta(self):
        m = {
            "low": self.low,
            "high": self.high,
        }
        return m


class Real(Parameter):
    """
    Real parameter class.

    Represents real parameters from R-space.

    Such parameters can be, for example, learning
    rate of ML oracle, or mutation rate in
    genetic oracles.

    Example:

    .. code-block:: python

        from feijoa.search.parameters import Real

        param = Real("foo", low=0.0, high=10.0)

    Args:
        name (str):
            Name of parameter.
        low (float):
            Lower bound of parameter.
        high (float):
            Upper bound of parameter.

    .. note:
        In parameters creation strict-typing is used
        for input values.

    Raises:
        ParametersIncorrectInputValues:
            If input values is not valid real values.
        AnyError:
            If anything bad happens.

    """

    def __init__(self, name: str, *, low: float, high: float):
        super().__init__(name)

        if not isinstance(low, float) or not isinstance(high, float):
            raise ParametersIncorrectInputValues(
                "Low and High bounds for" " Float parameter" " must be floats."
            )

        if low > high:
            raise ParametersIncorrectInputValues(
                "High bound must" " be greater than low bound."
            )

        self.low = low
        self.high = high

    def __repr__(self):
        return f"Real('{self.name}', low={self.low}, high={self.high})"

    @property
    def bounds(self):
        return np.array([self.low, self.high])

    def accept(self, v):
        return v.visit_real(self)

    def seed(self):
        return self.low

    def is_primitive(self):
        return True

    def get_unit_value(self, value):
        return float(value - self.low) / float(self.high - self.low)

    @property
    def meta(self):
        m = {
            "low": self.low,
            "high": self.high,
        }
        return m


class Categorical(Parameter):
    """
    Categorical parameter class.

    Represents category parameters.

    Such parameters can be, activation function,
    or type of regression model in AutoML.

    Args:
        name (str):
            Name of parameter.
        choices (list):
            List of choices, for example: ```['rbf', 'sigmoid']```

    .. note:
        In parameters creation strict-typing is used
        for input values.

    Raises:
        ParametersIncorrectInputValues:
            If input values is empty or exists
            not unique values.
        AnyError:
            If anything bad happens.

    """

    def __init__(self, name: str, *, choices):
        super().__init__(name)

        if not choices:
            raise ParametersIncorrectInputValues(
                "Choices for Categorical" " parameter must be not empty."
            )
        if len(choices) != len(set(choices)):
            raise ParametersIncorrectInputValues(
                "Choices for Categorical" " parameter must unique."
            )
        self.choices = choices

    def __repr__(self):
        return f"Categorical('{self.name}', choices={self.choices})"

    @property
    def bounds(self):
        return np.array([0, len(self.choices) - 1])

    def accept(self, v):
        return v.visit_categorical(self)

    def seed(self):
        return self.choices[0]

    def is_primitive(self):
        return False

    def get_unit_value(self, value):
        raise NotImplementedError()

    @property
    def meta(self):
        m = {"choices": self.choices}
        return m


class ParametersVisitor(metaclass=abc.ABCMeta):
    """
    Parameters abstract visitor

        Used for iteration over collections
        with parameters.

        Raises:
            AnyError:
                If anything bad happens.

    """

    @abc.abstractmethod
    def visit_integer(self, p: Integer):
        """Visit integer parameter."""

        raise NotImplementedError()

    @abc.abstractmethod
    def visit_real(self, p: Real):
        """Visit real parameter."""

        raise NotImplementedError()

    @abc.abstractmethod
    def visit_categorical(self, p: Categorical):
        """Visit categorical parameter."""

        raise NotImplementedError()



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import copy
import logging
from typing import Generator, List, Optional

from feijoa.models.configuration import Configuration
from feijoa.search.oracles.oracle import Oracle
from feijoa.search.parameters import Integer
from feijoa.search.space import SearchSpace
from feijoa.search.visitors import Randomizer

log = logging.getLogger(__name__)


class Pattern(Oracle):
    """
    Pattern search implementation.

    The result can be used for functions that are
    not its continuous or differentiable. One such
    search method is "convergence" (see below), based
    on the theory of positive reasons. The optimization
    turns into the best match (the solution with the
    smallest error value) in the multidimensional
    analytical possibility space.

    See more: https://en.wikipedia.org/wiki/Pattern_search_(optimization)

    Implementation based on: https://github.com/jansel/opentuner/blob/master/opentuner/search/patternsearch.py

    Raises:
        AnyError: If anything bad happens.

    """

    anchor = "template"
    aliases = (
        "PatternSearch",
        "pattern",
        "patternsearch",
        "TemplateSearch",
        "templatesearch",
    )

    def __init__(self, search_space: SearchSpace, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.search_space = search_space
        self.step_size = 0.1
        self.results: List[tuple] = []
        self._ask_gen: Optional[Generator] = None

    def ask(self, n: int = 1) -> Optional[List[Configuration]]:
        if not self._ask_gen:
            log.debug(
                "Parameter `n` does not affect on configuration's count,"
                f" because {self.__class__.__name__} is sequential."
            )
            self._ask_gen = self._ask(n)
        return next(self._ask_gen)

    def _ask(self, n: int) -> Generator:
        randomizer = Randomizer()
        center = {p.name: p.accept(randomizer) for p in self.search_space}
        yield [Configuration(center, requestor=self.name)]

        def set_unit_value(p, config, uv):
            """Set unit value."""

            low, high = p.low, p.high

            if isinstance(p, Integer):
                low -= 0.4999
                high += 0.4999

            if low < high:
                val = uv * float(high - low) + low

                if isinstance(p, Integer):
                    val = round(val)

                val = max(low, min(val, high))
                config[p.name] = val

        while True:
            points = list()

            for param in self.search_space:
                if param.is_primitive():

                    unit_value = param.get_unit_value(center[param.name])

                    if unit_value > 0.0:
                        down_cfg = copy.copy(center)
                        set_unit_value(
                            param,
                            down_cfg,
                            min(1.0, unit_value - self.step_size),
                        )
                        yield [Configuration(down_cfg, requestor=self.name)]
                        points.append(down_cfg)

                    if unit_value < 1.0:
                        up_cfg = copy.copy(center)
                        set_unit_value(
                            param,
                            up_cfg,
                            min(1.0, unit_value + self.step_size),
                        )
                        yield [Configuration(up_cfg, requestor=self.name)]
                        points.append(up_cfg)

                else:
                    cfg = copy.copy(center)
                    cfg[param.name] = param.accept(randomizer)
                    yield [Configuration(cfg, requestor=self.name)]
                    points.append(cfg)

            minima = min(self.results, key=lambda x: x[1])

            if minima[0] != center:
                center = copy.copy(minima[0])
            else:
                self.step_size /= 2.0

    def tell(self, config, result):
        self.results.append((config, result))



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Bayesian optimization module."""

import inspect
import logging
from typing import Generator, List, Optional

import numpy as np
from scipy.stats import norm

# noinspection PyUnresolvedReferences
# noinspection PyUnresolvedReferences
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.gaussian_process import GaussianProcessRegressor

from feijoa.models.configuration import Configuration
from feijoa.search.oracles.oracle import Oracle
from feijoa.search.visitors import Randomizer
from feijoa.utils.transformers import inverse_transform, transform

__all__ = ["Bayesian", "acquisition"]

log = logging.getLogger(__name__)


# noinspection PyPep8Naming
def acquisition(model, kind, X_samples, X, y, random_state=None, **kwargs):
    """
    Acquisition function for bayesian optimization.

    Args:
        model:
            Regressor model, must be fitted.
            Typically, Gaussian
        kind:
            Kind of acquisition function.
            Choices: ei, poi, ucb (only for GPR),
            lfboei, lfbopi, naive0 - model free.
        X_samples (numpy.ndarray):
            X samples to predict.
        X (numpy.ndarray):
            Features matrix.
        y (numpy.ndarray):
            Target values for X.
        random_state (int | None):
            Random state seed.

    Returns:
        Value of acquisition function.

    """

    if kind == "naive0":
        # experimental, research only
        return model.predict(X_samples)

    if kind == "lfboei":
        # https://arxiv.org/abs/2206.13035

        tau = np.quantile(y, q=0.33)
        classified = np.greater_equal(y, tau)
        x1, z1 = X[classified], classified[classified]
        x0, z0 = X, np.zeros_like(classified)
        w1 = (tau - y)[classified]
        w1 = w1 / np.mean(w1)
        w0 = 1 - z0

        x = np.concatenate([x1, x0], axis=0)
        z = np.concatenate([z1, z0], axis=0)

        s1 = x1.shape[0]
        s0 = x0.shape[0]

        # FIXME (qnbhd): ValueError: Input sample_weight contains NaN.
        w = np.concatenate(
            [w1 * (s1 + s0) / s1, w0 * (s1 + s0) / s0],
            axis=0,  # pragma: no mutate
        )
        w = w / np.mean(w)

        clf = RandomForestClassifier(
            n_jobs=-1, random_state=random_state
        )  # pragma: no mutate
        clf.fit(x, z, sample_weight=w)
        y_pred = clf.predict_proba(X_samples)[:, 1]
        return y_pred

    if kind == "lfbopoi":
        # https://arxiv.org/abs/2206.13035

        tau = np.quantile(y, q=0.33)
        classified = np.greater_equal(y, tau)
        clf = RandomForestClassifier(n_jobs=-1, random_state=random_state)
        clf.fit(X, classified)
        # FIXME (qnbhd): IndexError: index 1 is out of bounds for axis 1 with size 1
        y_pred = clf.predict_proba(X_samples)[:, 1]
        return y_pred

    assert isinstance(model, GaussianProcessRegressor), (
        "Model must be sklearn.GaussianProcessRegressor"
        "for classical ei, poi, ucb acquisition functions."
    )

    mean, std = model.predict(X_samples, return_std=True)
    best = min(mean)
    kappa = 2.5

    # regular acquisition functions.

    if kind == "poi":
        probs = norm.cdf((mean - best) / std)
        return probs

    elif kind == "ei":
        a = mean - best
        z = a / std
        return a * norm.cdf(z) + std * norm.pdf(z)

    elif kind == "ucb":
        return mean + kappa * std

    raise ValueError("Not correct acquisition function kind.")


class Bayesian(Oracle):
    """
    Bayesian optimization oracle.

    Bayesian optimization is a global optimization
    method for an unknown function (black box) with noise.
    Bayesian optimization applied to hyperparameters
    optimization builds a stochastic model of the mapping
    function from hyperparameter value to an objective
    function applied on the test set. By iteratively
    applying a perspective hyperparameter configuration
    based on the current model and then updating it,
    Bayesian optimization seeks to gather as much
    information as possible about that function and,
    in particular, the location of the optimum. The method
    attempts to balance probing (hyperparameters for which
    the change is least reliably known) and exploitation
    (hyperparameters that are expected to be closest to
    the optimum). In practice, Bayesian optimization
    has shown better results with less computation compared
    to grid search and random search due to the ability
    to judge the quality of experiments even before they
    are performed.

    Args:
        search_space (SearchSpace):
            Search space instance.
        acq (str):
            Acquisition function.
            Can be `pi`, `ucb`, `ei` with Gaussian Regressor
            Or `naive0` - experimental, `lfboei`, `lfbopi`
        regr:
            Regression model, must have
                - fit(X, y)
                - predict_proba(X)
                - predict(X)
            methods. For example, you
            can use regressor from sklearn package
            (https://scikit-learn.org/)
        n_warmup (int):
            Warmup points count.

    .. note::
        `lfbopi` and `lfboei` taked from publication:
         https://arxiv.org/abs/2206.13035

    .. note::
        `naive0` - experimental function.

    """

    anchor = "bayesian"

    aliases = (
        "BayesianOracle",
        "bayesian",
        "bayes",
    )

    def __init__(
        self,
        search_space,
        *args,
        acq="ei",
        seed=0,
        regr="GaussianProcessRegressor",
        n_warmup=5,
        plugins=None,
        **kwargs,
    ):

        super().__init__(*args, **kwargs)

        for p in plugins or []:
            # TODO (qnbhd): add task case with plugin.
            self.attach(p)

        self.search_space = search_space

        # build regression model

        # TODO (qnbhd): elliminate globals
        regr = globals()[regr]

        self.model = regr() if inspect.isclass(regr) else regr

        # take bounds

        self.bounds = search_space.bounds

        # trials pool for fitting

        self.X = np.empty(shape=(0, len(self.search_space)))

        # results pool for fitting

        self.y = np.empty(shape=(0,))

        # setup acquisition function

        if not isinstance(self.model, GaussianProcessRegressor):
            choices = ["lfboei", "lfbopoi", "naive0"]
            if acq in choices:
                self.acq_function = acq
            else:
                raise ValueError(
                    f"Acquisition function `{acq}` not"
                    f" available with `{regr}` regressor."
                )
        else:
            self.acq_function = acq

        log.critical(f"ACQ function: {self.acq_function}")

        # warmup points
        self.n_warmup = n_warmup

        # setup readable name
        self._name = f"Bayesian<{type(self.model).__name__}({self.acq_function})>"

        self._ask_gen = None

    def ask(self, n: int = 1) -> Optional[List[Configuration]]:
        # TODO (qnbhd): add test case on configurations
        #  per emit count

        if not self._ask_gen:
            # build ask generator
            self._ask_gen = self._ask(n)

        return next(self._ask_gen)

    def _ask(self, n: int) -> Generator:
        """Main ask generator."""

        # make some warmup configurations

        randomizer = Randomizer(self.seed)
        random_samples = [
            Configuration(
                {p.name: p.accept(randomizer) for p in self.search_space},
                requestor=self.name,
            )
            for _ in range(self.n_warmup)
        ]

        yield random_samples

        self.model.fit(self.X, self.y)

        while True:
            x = self.opt_acquisition(n)
            configurations = list()
            for c in x:
                configurations.append(
                    Configuration(
                        transform(c, self.search_space),
                        requestor=self.name,
                    )
                )
            yield configurations
            self.model.fit(self.X, self.y)

    def opt_acquisition(self, n: int):
        """Optimize acquisition function."""

        n_samples = 100000

        X_samples = np.random.uniform(
            self.bounds[:, 0],
            self.bounds[:, 1],
            size=(n_samples, self.bounds.shape[0]),
        )

        scores = acquisition(
            self.model,
            self.acq_function,
            X_samples,
            self.X,
            self.y,
            random_state=self.seed,
        )

        assert len(scores) == n_samples

        minima_x = X_samples[scores.argsort()[:n]]

        return minima_x

    def tell(self, config, result):
        """Tell configuration's result."""

        vec = np.array(inverse_transform(config, self.search_space))

        self.X = np.concatenate([self.X, vec.reshape(1, -1)])  # pragma: no mutate
        self.y = np.concatenate([self.y, [result]])

        if len(self.y) > 5:  # pragma: no mutate
            self.notify("on_tell", config, result)

    def update(self, event, subject, *args, **kwargs):
        log.info(
            f"Event: {event}," f"Subject: {subject}" f"Args: {args}" f"Kwargs: {kwargs}"
        )



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""TinyDB storage module."""

from typing import List, NamedTuple, Optional

from feijoa.utils.imports import ImportWrapper

with ImportWrapper():
    from tinydb import Query, TinyDB
    from tinydb.table import Document
    from tinydb.operations import set as tinydb_set

from feijoa.exceptions import DBVersionError, InsertExperimentWithTheExistedId
from feijoa.models import Experiment
from feijoa.models.configuration import Configuration
from feijoa.search.space import SearchSpace
from feijoa.storages.storage import Storage

__all__ = ["TinyDBStorage"]


class _Parameter(NamedTuple):
    name: str
    kind: str
    meta: dict


# noinspection PyTypeChecker
class TinyDBStorage(Storage):
    """TinyDB storage uses tinydb library."""

    __version__ = "0.1.0"

    def __init__(self, json_file: str):
        self.json_file = json_file
        self.tiny_db = TinyDB(json_file)

        db_version_table = self.tiny_db.table("version")

        if not len(db_version_table.all()):
            # New table
            db_version_table.insert({"version": self.version})
        else:
            ver_record, *_ = db_version_table.all()
            ver = ver_record["version"]
            if ver < self.version:
                raise DBVersionError()

        self.jobs_table = self.tiny_db.table("job")
        self.experiments_table = self.tiny_db.table("experiment")
        self.parameters_table = self.tiny_db.table("parameters")

    def insert_job(self, job):
        doc = {
            "name": job.name,
            "id": job.id,
            "last_optimizer": job.optimizer_name_dsl,
        }

        self.jobs_table.insert(doc)

        for p in job.search_space:
            name = p.name
            kind = p.__class__.__name__
            meta = p.meta
            param_model = {
                "job_id": job.id,
                "name": name,
                "kind": kind,
                "meta": meta,
            }
            self.parameters_table.insert(param_model)

    def update_optimizer_name_by_job_id(self, job_id, name):
        self.jobs_table.update(tinydb_set("last_optimizer", name), Query().id == "John")

    def get_search_space_by_job_id(self, job_id) -> SearchSpace:
        parameters = self.parameters_table.search(Query().job_id == job_id)
        pool = list()
        for p in parameters:
            mod = _Parameter(name=p["name"], kind=p["kind"], meta=p["meta"])
            pool.append(mod)
        return SearchSpace.from_db_parameters(pool)

    def get_optimizer_name_by_job_id(self, job_id) -> Optional[str]:
        jobs = self.jobs_table.search(Query().id == job_id)

        if not len(jobs):
            return None

        assert len(jobs) == 1

        job, *_ = jobs

        return job["optimizer_name_dsl"]

    def get_job_id_by_name(self, name) -> Optional[int]:
        jobs = self.jobs_table.search(Query().name == name)

        if not len(jobs):
            return None

        assert len(jobs) == 1

        job, *_ = jobs

        return job["id"]

    def is_job_name_exists(self, name):
        jobs = self.jobs_table.search(Query().name == name)
        return len(jobs) != 0

    def insert_experiment(self, experiment):
        doc = experiment.dict()
        doc["requestor"] = experiment.params.requestor

        if self.get_experiment(experiment.job_id, experiment.id):
            # TODO (qnbhd): make correct exception
            raise InsertExperimentWithTheExistedId()

        self.experiments_table.insert(doc)

    def get_experiment(self, job_id, experiment_id):
        q = self.experiments_table.search(
            (Query().id == experiment_id) & (Query().job_id == job_id)
        )

        if not q:
            return None

        exp = q[0]
        exp["params"] = Configuration(exp["params"], requestor=exp["requestor"])
        exp.pop("requestor")

        return Experiment(**exp)

    def _get_raw_experiments(self, job_id) -> List[Document]:
        docs = self.experiments_table.search(Query().job_id == job_id)
        return docs

    def get_experiments_by_job_id(self, job_id) -> List[Experiment]:
        docs = self._get_raw_experiments(job_id)
        experiments = []
        for doc in docs:
            doc["params"] = Configuration(doc["params"], requestor=doc["requestor"])
            # doc.pop('requestor')
            exp = Experiment(**doc)
            experiments.append(exp)
        return experiments

    def get_experiments_count(self, job_id) -> int:
        return len(self.get_experiments_by_job_id(job_id))

    @property
    def jobs(self):
        return self.jobs_table.all()

    @property
    def version(self):
        return self.__version__

    def __del__(self):
        self.tiny_db.close()



import logging
import math
import sys
from typing import Generator, List, Optional, Tuple

import numpy as np
import pandas as pd
from scipy.stats import multivariate_normal

from feijoa.models.configuration import Configuration
from feijoa.search.oracles.oracle import Oracle
from feijoa.utils.transformers import transform


# noinspection PyAbstractClass,SpellCheckingInspection,PyPep8Naming
class BCMAES(Oracle):

    anchor: str = "BCMAES"
    aliases: Tuple = ("bcmaes", "BCMAES", "bCMAES")

    def __init__(self, search_space, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.list_x_star = list()
        self.list_fx_star = list()

        self.space = search_space
        self.bounds = self.space.bounds

        # TODO (qnbhd): change to
        # n = 4 + int(3 * math.log(len(mu_0)))

        self.mu0 = np.random.uniform(
            self.bounds[:, 0],
            self.bounds[:, 1],
            size=(self.bounds.shape[0],),
        )

        self.tol = 0.005

        self.n = 4 + int(3 * math.log(len(self.mu0)))
        self.k0 = 4
        self.v0 = self.n + 2
        self.factor = 1
        self.psi = np.eye(2)
        self.x_star_min = self.mu0
        self.var_star_min = self.psi
        self.fx_star_min = sys.float_info.max
        self.count = 0
        self.step = 0

        self.variance = np.ones([self.n, self.n])
        self.results = []

        self._ask_gen = None

    def ask(self, n: int = 1) -> Optional[List[Configuration]]:
        # TODO (qnbhd): add test case on configurations
        #  per emit count

        if not self._ask_gen:
            # build ask generator
            self._ask_gen = self._ask(n)

        return next(self._ask_gen)

    # noinspection PyUnusedLocal
    def _ask(self, W: int) -> Generator:
        while True:
            # while np.linalg.norm(self.variance) > self.tol:
            E_mu = self.mu0
            E_sigma = self.psi / (self.v0 - self.n - 1)
            self.step += 1

            # FIXME (qnbhd): ValueError: mean and cov must have same length
            sample_X = np.random.multivariate_normal(
                mean=E_mu, cov=E_sigma, size=(self.n,)
            )
            sample_X = np.clip(sample_X, self.bounds[:, 0], self.bounds[:, 1])

            yield self.make_configs(sample_X)

            # TODO (qnbhd): FIX
            g_x = self.results.pop()

            df = pd.DataFrame(sample_X)
            df["g_x"] = g_x
            df["g_x"] = g_x
            df["d"] = multivariate_normal.pdf(
                sample_X, mean=E_mu, cov=E_sigma, allow_singular=True
            )
            x_order_d = df.sort_values(by=["d"], ascending=False)
            x_order_f = x_order_d.sort_values(by=["g_x"], kind="mergesort")

            d_ordered = x_order_d[x_order_d.columns[-1:]].values / sum(df["d"])
            x_order_f = x_order_f[x_order_f.columns[:-2]].values
            x_order_d = x_order_d[x_order_d.columns[:-2]].values

            x_all = df[df.columns[:-2]].values
            d = np.array(df["d"])[:, None] / sum(df["d"])
            x_bar_f = np.sum(x_order_f * d_ordered, axis=0)
            x_bar = np.sum(x_all * d, axis=0)
            mean = x_order_f[0]

            # sigma part :
            sigma_emp = np.dot((x_all - x_bar).T, (x_all - x_bar) * d)
            sigma_ordered = np.dot(
                (x_order_f - x_bar_f).T,
                (x_order_f - x_bar_f) * d_ordered,
            )
            # other :
            self.variance = (sigma_ordered - (sigma_emp - E_sigma)) * self.factor

            variance_norm = np.linalg.norm(self.variance)

            # TODO (qnbhd): FIX
            yield self.make_configs([mean])

            fx_star = self.results.pop()

            if fx_star < self.fx_star_min:
                # if self.fx_star_min - fx_star < 1e-5:
                #     return list_x_star, list_fx_star
                self.count = 0
                self.factor = 1
                self.fx_star_min = fx_star
                self.x_star_min = mean
                if variance_norm < 100 * self.n:
                    self.var_star_min = self.variance
                self.list_x_star.append(mean)

                yield self.make_configs([mean])

                # TODO (qnbhd): FIX
                self.list_fx_star.append(self.results.pop())
            else:
                self.list_x_star.append(self.x_star_min)
                # TODO (qnbhd): FIX
                yield self.make_configs([self.x_star_min])
                self.list_fx_star.append(self.results.pop())

                self.count += 1

                if variance_norm > 100 * self.n:
                    mean = self.x_star_min
                    self.variance = self.var_star_min
                    self.mu0 = mean

                if self.count > 5:
                    self.factor = 1.5
                if self.count == 20:
                    mean = self.x_star_min
                    self.variance = self.var_star_min
                    self.mu0 = mean
                if self.count > 20:
                    self.factor = 0.9
                if self.count > 30:
                    self.factor = 0.7
                if self.count > 40:
                    self.factor = 0.5

            self.mu0 = (self.mu0 * self.k0 + self.n * mean) / (self.k0 + self.n)
            self.k0 = self.k0 + self.n
            self.v0 = self.v0 + self.n
            self.psi = (
                self.psi
                + (self.k0 * self.n)
                / (self.k0 + self.n)
                * np.dot(np.mat(mean - self.mu0), np.mat(mean - self.mu0).T)
                + self.variance * (self.n - 1)
            )
            self.psi = self.psi * self.factor

    def tell(self, config, result):
        self.results.append(result)

    def make_configs(self, sample):
        configurations = list()
        for c in sample:
            configurations.append(
                Configuration(
                    # transform(np.clip(c, self.bounds[:, 0], self.bounds[:, 1]), self.space),
                    transform(c, self.space),
                    requestor=self.name,
                )
            )
        return configurations



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import numpy as np

from feijoa.search.parameters import Categorical, Integer, Real


def transform(sol, search_space):
    """
    Transform solutions into a dictionary.

        Args:
            sol (list | numpy.array):
                Vector of parameter's values.
            search_space:
                Search space instance.

        Raises:
            AnyError: If anything bad happens.

    """

    configuration = dict()

    for value, param in zip(sol, search_space):
        if isinstance(param, Real):
            configuration[param.name] = value
            continue

        rounded = int(value + (0.5 if value > 0 else -0.5))

        if isinstance(param, Integer):
            configuration[param.name] = rounded
        elif isinstance(param, Categorical):
            configuration[param.name] = param.choices[rounded]

    return configuration


def inverse_transform(configuration: dict, search_space):
    """
    Transform dictionary (configuration)
        to vector of parameter's values.

        Args:
            configuration (dict | Configuration):
                Feijoa configuration.
            search_space:
                Search space instance.

        Raises:
            AnyError: If anything bad happens.

    """

    solution = np.zeros((len(configuration),), dtype=np.float64)

    for i, (key, value) in enumerate(configuration.items()):
        param = search_space.get(key)

        if isinstance(param, (Real, Integer)):
            solution[i] = value

        if isinstance(param, Categorical):
            index = param.choices.index(value)
            solution[i] = index

    return solution



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Base plugin's class module."""

from abc import ABCMeta

from feijoa.utils.mixins import Observer, Subject

__all__ = ["Plugin"]


class Plugin(Observer, Subject, metaclass=ABCMeta):
    """
    Base plugin's class.

    The idea of plugins is to optionally add
    new functionality to search oracles.
    The plugin inherits from the Subject and
    Observer classes. This means that he can
    be both a subject (receive notification
    from oracle), as well as an observer
    (send alerts to oracle).

    Args:
        subscribers (list):
            Initial array of subscribers.

    Raises:
        AnyError: If anything bad happens.

    """

    @property
    def anchor(self):
        raise NotImplementedError()

    @property
    def aliases(self):
        raise NotImplementedError()

    def __init__(self, subscribers=None):
        self.subscribers = subscribers or list()

    def attach(self, observer: Observer) -> None:
        self.subscribers.append(observer)

    def detach(self, observer: Observer) -> None:
        self.subscribers.remove(observer)

    def notify(self, event, *args, **kwargs) -> None:
        for o in self.subscribers:
            o.update(event, self, *args, **kwargs)



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Meta oracle base class module."""

import abc
import logging
from typing import Generator, List, Optional

from feijoa.models.configuration import Configuration
from feijoa.search.oracles.oracle import Oracle
from feijoa.utils.imports import LazyModuleImportProxy

seed = LazyModuleImportProxy("feijoa.search.seed")


log = logging.getLogger(__name__)


class MetaOracle(Oracle):
    """
    Meta-search oracle.

    Used for pick up oracles.

    Raises:
        AnyError: If anything bad happens.

    """

    def __init__(self, *oracles, **kwargs):
        super().__init__(*oracles, **kwargs)
        self.oracles = list(oracles)
        self.ask_gen = None

    def ask(self, n: int = 1) -> Optional[List[Configuration]]:
        if not self.ask_gen:
            self.ask_gen = self._ask(n)
        return next(self.ask_gen)

    def _ask(self, n: int) -> Generator:
        prev_picked = ""

        # pick up all seed's oracles

        seed_oracles = [a for a in self.oracles if isinstance(a, seed.SeedOracle)]

        # remove all seed's oracles

        for oracle in seed_oracles:
            self.remove_oracle(oracle)

        if seed_oracles:
            log.debug("Let's measure seed configurations!")
            for oracle in seed_oracles:
                configs = oracle.ask(n)
                yield configs

        while True:
            for oracle in self.order:
                if not prev_picked or prev_picked != oracle.name:
                    log.debug(f"Pick {oracle.name}")
                configs = oracle.ask(n)
                yield configs
                prev_picked = oracle.name

    def tell(self, config, result):
        """
        Tell results to all search oracles"""

        for oracle in self.oracles:
            oracle.tell(config, result)

    def add_oracle(self, oracle: Oracle):
        """
        Append oracle to oracles list.

        Args:
            oracle (Oracle):
                search oracle instance.

        Returns:
            None

        Raises:
            AnyError: If anything bad happens.

        """

        self.oracles.append(oracle)

    @abc.abstractmethod
    def remove_oracle(self, oracle: Oracle):
        """
        Remove oracle from inner oracles
        pool in meta-oracle.

        Args:
            oracle (Oracle):
                search oracle instance.

        Returns:
            None

        Raises:
            AnyError: If anything bad happens.

        """

    @property
    @abc.abstractmethod
    def order(self):
        """Get order for oracles."""

        raise NotImplementedError()



# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = "3.10"

_lr_method = "LALR"

_lr_signature = "ANGLEBRACKETL ANGLEBRACKETR COMMA EQ FLOAT INTEGER PLUS SQBRACKETL SQBRACKETR WORD\n        super_oracle : WORD ANGLEBRACKETL oracle-list ANGLEBRACKETR\n                     | WORD ANGLEBRACKETL oracle-list ANGLEBRACKETR SQBRACKETL params-list SQBRACKETR\n                     | oracle\n        \n        oracle : WORD\n               | WORD SQBRACKETL params-list SQBRACKETR\n               | WORD PLUS plugins-list\n               | WORD SQBRACKETL params-list SQBRACKETR PLUS plugins-list\n        \n        plugin : WORD\n               | WORD SQBRACKETL params-list SQBRACKETR\n        \n        plugins-list : plugin\n                     | plugins-list PLUS plugin\n        \n        oracle-list : oracle\n                    | oracle-list COMMA oracle\n        \n        value : INTEGER\n              | FLOAT\n              | WORD\n        param : WORD EQ value\n        params-list : param\n                    | params-list COMMA param\n        "

_lr_action_items = {
    "WORD": (
        [
            0,
            4,
            5,
            6,
            17,
            18,
            20,
            21,
            22,
            23,
            29,
        ],
        [
            2,
            7,
            10,
            13,
            7,
            25,
            10,
            10,
            13,
            10,
            13,
        ],
    ),
    "$end": (
        [
            1,
            2,
            3,
            13,
            14,
            15,
            16,
            19,
            32,
            34,
            35,
            36,
        ],
        [
            0,
            -4,
            -3,
            -8,
            -6,
            -10,
            -1,
            -5,
            -11,
            -7,
            -9,
            -2,
        ],
    ),
    "ANGLEBRACKETL": (
        [
            2,
        ],
        [
            4,
        ],
    ),
    "SQBRACKETL": (
        [
            2,
            7,
            13,
            16,
        ],
        [
            5,
            5,
            21,
            23,
        ],
    ),
    "PLUS": (
        [
            2,
            7,
            13,
            14,
            15,
            19,
            32,
            34,
            35,
        ],
        [
            6,
            6,
            -8,
            22,
            -10,
            29,
            -11,
            22,
            -9,
        ],
    ),
    "ANGLEBRACKETR": (
        [
            7,
            8,
            9,
            13,
            14,
            15,
            19,
            24,
            32,
            34,
            35,
        ],
        [
            -4,
            16,
            -12,
            -8,
            -6,
            -10,
            -5,
            -13,
            -11,
            -7,
            -9,
        ],
    ),
    "COMMA": (
        [
            7,
            8,
            9,
            11,
            12,
            13,
            14,
            15,
            19,
            24,
            25,
            26,
            27,
            28,
            30,
            31,
            32,
            33,
            34,
            35,
        ],
        [
            -4,
            17,
            -12,
            20,
            -18,
            -8,
            -6,
            -10,
            -5,
            -13,
            -16,
            -17,
            -14,
            -15,
            -19,
            20,
            -11,
            20,
            -7,
            -9,
        ],
    ),
    "EQ": (
        [
            10,
        ],
        [
            18,
        ],
    ),
    "SQBRACKETR": (
        [
            11,
            12,
            25,
            26,
            27,
            28,
            30,
            31,
            33,
        ],
        [
            19,
            -18,
            -16,
            -17,
            -14,
            -15,
            -19,
            35,
            36,
        ],
    ),
    "INTEGER": (
        [
            18,
        ],
        [
            27,
        ],
    ),
    "FLOAT": (
        [
            18,
        ],
        [
            28,
        ],
    ),
}

_lr_action = {}
for _k, _v in _lr_action_items.items():
    for _x, _y in zip(_v[0], _v[1]):
        if not _x in _lr_action:
            _lr_action[_x] = {}
        _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {
    "super_oracle": (
        [
            0,
        ],
        [
            1,
        ],
    ),
    "oracle": (
        [
            0,
            4,
            17,
        ],
        [
            3,
            9,
            24,
        ],
    ),
    "oracle-list": (
        [
            4,
        ],
        [
            8,
        ],
    ),
    "params-list": (
        [
            5,
            21,
            23,
        ],
        [
            11,
            31,
            33,
        ],
    ),
    "param": (
        [
            5,
            20,
            21,
            23,
        ],
        [
            12,
            30,
            12,
            12,
        ],
    ),
    "plugins-list": (
        [
            6,
            29,
        ],
        [
            14,
            34,
        ],
    ),
    "plugin": (
        [
            6,
            22,
            29,
        ],
        [
            15,
            32,
            15,
        ],
    ),
    "value": (
        [
            18,
        ],
        [
            26,
        ],
    ),
}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
    for _x, _y in zip(_v[0], _v[1]):
        if not _x in _lr_goto:
            _lr_goto[_x] = {}
        _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
    ("S' -> super_oracle", "S'", 1, None, None, None),
    (
        "super_oracle -> WORD ANGLEBRACKETL oracle-list ANGLEBRACKETR",
        "super_oracle",
        4,
        "p_super_oracle",
        "parser.py",
        78,
    ),
    (
        "super_oracle -> WORD ANGLEBRACKETL oracle-list ANGLEBRACKETR SQBRACKETL params-list SQBRACKETR",
        "super_oracle",
        7,
        "p_super_oracle",
        "parser.py",
        79,
    ),
    (
        "super_oracle -> oracle",
        "super_oracle",
        1,
        "p_super_oracle",
        "parser.py",
        80,
    ),
    ("oracle -> WORD", "oracle", 1, "p_oracle", "parser.py", 97),
    (
        "oracle -> WORD SQBRACKETL params-list SQBRACKETR",
        "oracle",
        4,
        "p_oracle",
        "parser.py",
        98,
    ),
    (
        "oracle -> WORD PLUS plugins-list",
        "oracle",
        3,
        "p_oracle",
        "parser.py",
        99,
    ),
    (
        "oracle -> WORD SQBRACKETL params-list SQBRACKETR PLUS plugins-list",
        "oracle",
        6,
        "p_oracle",
        "parser.py",
        100,
    ),
    ("plugin -> WORD", "plugin", 1, "p_plugin", "parser.py", 114),
    (
        "plugin -> WORD SQBRACKETL params-list SQBRACKETR",
        "plugin",
        4,
        "p_plugin",
        "parser.py",
        115,
    ),
    (
        "plugins-list -> plugin",
        "plugins-list",
        1,
        "p_plugins_list",
        "parser.py",
        125,
    ),
    (
        "plugins-list -> plugins-list PLUS plugin",
        "plugins-list",
        3,
        "p_plugins_list",
        "parser.py",
        126,
    ),
    (
        "oracle-list -> oracle",
        "oracle-list",
        1,
        "p_oracle_list",
        "parser.py",
        134,
    ),
    (
        "oracle-list -> oracle-list COMMA oracle",
        "oracle-list",
        3,
        "p_oracle_list",
        "parser.py",
        135,
    ),
    ("value -> INTEGER", "value", 1, "p_value", "parser.py", 157),
    ("value -> FLOAT", "value", 1, "p_value", "parser.py", 158),
    ("value -> WORD", "value", 1, "p_value", "parser.py", 159),
    (
        "param -> WORD EQ value",
        "param",
        3,
        "p_param",
        "parser.py",
        165,
    ),
    (
        "params-list -> param",
        "params-list",
        1,
        "p_params_list",
        "parser.py",
        171,
    ),
    (
        "params-list -> params-list COMMA param",
        "params-list",
        3,
        "p_params_list",
        "parser.py",
        172,
    ),
]



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Histogram's plot module."""

import plotly.graph_objs as go


def plot_objective_hist(
    job,
    fig=None,
    invert_objective=False,
):
    """
    Plot objective hists for
    a specified job.

    Args:
        job (Job):
            Job instance.
        fig (go.Figure):
            Plotly figure object.
        invert_objective (bool):
            For the maximization task, it is
            necessary to invert the value of
            the view function, in this case,
            you must check this box to get
            corrective values

    Raises:
        AnyError: If anything bad happens.

    """

    fig = fig or go.Figure()
    df = job.get_dataframe(brief=True)
    objectives = df["objective_result"] * (1 if not invert_objective else -1)

    fig.add_histogram(x=objectives, nbinsx=100)

    fig.update_layout(
        title="Objective result histogram",
        xaxis_title="Objective value",
        yaxis_title="Count",
    )

    return fig



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""EDF plot module."""

import numpy as np
import plotly.graph_objs as go

__all__ = ["plot_edf"]


def plot_edf(
    job,
    fig=None,
    name=None,
    invert_objective=False,
):
    """
    Plot the EDF of a job.

    Args:
        job (Job):
            Job instance.
        fig (go.Figure):
            Plotly figure object.
        name:
            Name of trace.
        invert_objective (bool):
            For the maximization task, it is
            necessary to invert the value of
            the view function, in this case,
            you must check this box to get
            corrective values

    Raises:
        AnyError: If anything bad happens.

    """

    name = name or job.name
    fig = fig or go.Figure()
    df = job.get_dataframe(brief=True)
    objectives = df["objective_result"] * (1 if not invert_objective else -1)
    min_value, max_value = objectives.min(), objectives.max()
    lspace = np.linspace(min_value, max_value, 100)

    dist = np.array([np.sum(objectives <= x) for x in lspace]) / len(objectives)

    fig.add_trace(
        go.Scatter(
            x=lspace,
            y=dist,
            mode="lines",
            name=f"{name}",
        )
    )

    fig.update_layout(
        title="Empirical distribution plot",
        xaxis_title="Objective value",
        yaxis_title="Cumulative probability",
    )

    return fig



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Grid search class module."""

import logging
from itertools import product
from typing import Generator, List, Optional

import numpy

from feijoa.models.configuration import Configuration
from feijoa.search.oracles.oracle import Oracle
from feijoa.search.parameters import Categorical, Integer, ParametersVisitor, Real

__all__ = ["Grid"]

log = logging.getLogger(__name__)


# noinspection PyUnresolvedReferences
class GridMaker(ParametersVisitor):
    """
    Simple grid creation for different parameters.

    Args:
        EPS (float):
            Sample rate for included-valued parameters

    Raises:
        AnyError: If anything bad happens.

    """

    EPS = 0.1

    def visit_integer(self, p: Integer):
        """Pick up integer grid."""

        return range(p.low, p.high + 1)

    def visit_real(self, p: Real):
        """Pick up real grid."""

        return numpy.round(
            numpy.arange(p.low, p.high + GridMaker.EPS, GridMaker.EPS),
            2,
        )

    def visit_categorical(self, p: Categorical):
        """Pick up categorical grid."""

        return p.choices


class Grid(Oracle):
    """
    Simple grid search.

    Makes a full iteration over a manually
    specified subset of hyperparameter spaces
    of the oracle.

    Raises:
        AnyError: If anything bad happens.

    """

    anchor = "grid"
    aliases = (
        "GridSearch",
        "grid",
    )

    def __init__(self, search_space, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.search_space = search_space

        # grids for all parameters

        grids = [p.accept(GridMaker()) for p in self.search_space.params]

        # iterator for grid search

        self.prod_iter = product(*grids)

        self._ask_gen = None

    def ask(self, n: int = 1) -> Optional[List[Configuration]]:
        if not self._ask_gen:
            self._ask_gen = self._ask(n)
        return next(self._ask_gen)

    def _ask(self, n: int) -> Generator:

        while True:
            cfgs = []

            for _ in range(n):
                try:
                    generation = next(self.prod_iter)
                except StopIteration:
                    break

                cfg = {h.name: v for h, v in zip(self.search_space, generation)}

                cfgs.append(Configuration(cfg, requestor=self.name))

            if not cfgs:
                yield None

            yield cfgs

    def tell(self, config, result):
        """No needed."""

        pass



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Base class of search oracles."""

import abc
import random
from typing import List, Optional

import numpy

from feijoa.models.configuration import Configuration

__all__ = ["Oracle"]

from feijoa.utils.mixins import Observer, Subject


class Oracle(Subject, Observer, metaclass=abc.ABCMeta):
    """
    Base class of search oracles.

    Have a simple ask-tell interface.

    When creating the possibility of implementing
    oracles, it is desirable that the constructor
    arguments should not contain any dependencies on
    feijoa, except for SearchSpace. This decision is
    made to keep the oracles easy to use without
    initializing feijoa components.

    All oracles must implement:

        anchor:
            Name of oracle, which
            used for dynamic searching all oracles
            in project.

        aliases:
            synonyms for oracle name.

        ask(n):
            Get k <= n configurations.

        tell(configuration, result):
            Tell oracle measured result for
            configuration


    Raises:
        AnyError: If anything bad happens.

    """

    def __init__(self, *args, seed=0, **kwargs):
        self._name = self.__class__.__name__
        self.subscribers = []
        numpy.random.seed(seed)
        random.seed(seed)
        self.seed = seed

    @property
    @abc.abstractmethod
    def anchor(self):
        """Name used for dynamic oracle detection."""

        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def aliases(self):
        """Synonyms for oracle name."""

        raise NotImplementedError()

    @abc.abstractmethod
    def ask(self, n: int = 1) -> Optional[List[Configuration]]:
        """
        Get configurations from oracle.

        Args:
            n (int, optional):
                Preferred count of configurations.

        .. note::
            Parameter **n** may not affect on
            configuration's count.

        Returns:
            List of configurations, or None if
            configurations are over.

        Raises:
            AnyError: If anything bad happens.

        """

        raise NotImplementedError()

    @abc.abstractmethod
    def tell(self, config, result):
        """
        Tell configuration's result to oracle
        to accumulate information model.

        Args:
            config (Configuration):
                Configuration instance.
            result (float):
                Objective result for configuration.

        .. note::
            The function signature may change in the
            future due to plans to implement
            multi&many-objective optimization.

        Returns:
            None

        Raises:
            AnyError: If anything bad happens.

        """

        raise NotImplementedError()

    @property
    def name(self):
        """Name of oracle."""

        return self._name

    @name.setter
    def name(self, v):
        """Name setter for oracle."""

        self._name = v

    def attach(self, observer: Observer) -> None:
        self.subscribers.append(observer)

    def detach(self, observer: Observer) -> None:
        self.subscribers.remove(observer)

    def notify(self, event, *args, **kwargs):
        for observer in self.subscribers:
            observer.update(event, self, *args, **kwargs)

    def update(self, event, subject, *args, **kwargs):
        pass



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Reducer class module."""

import logging

import numpy as np

from feijoa.plugins.plugin import Plugin
from feijoa.search.parameters import Categorical, Integer, Real
from feijoa.search.space import SearchSpace

log = logging.getLogger(__name__)

__all__ = ["DomainReducer"]


class DomainReducer(Plugin):
    """
    A sequential domain reduction transformer bassed
    on the work by Stander, N. and Craig, K:
    "On the robustness of a simple domain reduction
    scheme for simulation‐based optimization"

    Link: http://www.truegrid.com/srsm_revised.pdf

    Based on: https://github.com/fmfn/BayesianOptimization

    Args:
        gamma_osc (float):
            Shrinkage parameter for oscillation.
            Typically, [0.5-0.7].
        gamma_pan (float):
            Panning parameter. Typically, 1.0.
        eta (float):
            Zoom parameter.

    Raises:
        AnyError: If anything bad happens.

    """

    anchor = "reducer"
    aliases = ("domain_reducer",)

    def __init__(
        self,
        gamma_osc: float = 0.7,
        gamma_pan: float = 1.2,
        eta: float = 0.9,
        min_window_value: float = 0.0,
        subscribers=None,
    ):

        super().__init__(subscribers=subscribers)

        self.gamma_osc = gamma_osc
        self.gamma_pan = gamma_pan
        self.eta = eta

        self.min_window_value = min_window_value
        self.min_window: np.ndarray = np.array([])

        self.bounds: np.ndarray = np.array([])
        self.original_bounds: np.ndarray = np.array([])

        self.previous_optimal: np.ndarray = np.array([])
        self.current_optimal: np.ndarray = np.array([])

        self.previous_d: np.ndarray = np.array([])
        self.current_d: np.ndarray = np.array([])

        self.c_hat: np.ndarray = np.array([])
        self.r: np.ndarray = np.array([])
        self.contraction_rate: float = 0.0
        self.c: np.ndarray = np.array([])
        self.gamma: np.ndarray = np.array([])

        self.fitted = False

    def init(self, space):
        """
        Init reducer.

        Args:
            space (SearchSpace):
                Search space instance.

        Raises:
            AnyError: If anything bad happens.

        """

        bounds = []

        for p in space:
            if isinstance(p, (Integer, Real)):
                bounds.append((p.low, p.high))
            if isinstance(p, Categorical):
                bounds.append((0, len(p.choices) - 1))

        self.original_bounds = np.array(bounds)
        self.bounds = np.array(np.copy(self.original_bounds))
        self.min_window = np.array([self.min_window_value] * len(self.bounds))

        self.previous_optimal = np.mean(self.bounds, axis=1)
        self.current_optimal = np.mean(self.bounds, axis=1)

        self.r = self.original_bounds[:, 1] - self.original_bounds[:, 0]

        self.current_d = 2.0 * (self.current_optimal - self.previous_optimal) / self.r
        self.previous_d = np.copy(self.current_d)

    # noinspection PyPep8Naming
    def refit(self, x_opt: np.ndarray):
        self.previous_optimal = self.current_optimal
        self.previous_d = x_opt

        self.current_d = (
            2.0
            # pyre-ignore[58]:
            * (self.current_optimal - self.previous_optimal)
            / self.r
        )

        self.c = self.current_d * self.previous_d

        self.c_hat = np.sqrt(np.abs(self.c)) * np.sign(self.c)

        self.gamma = 0.5 * (
            self.gamma_pan * (1.0 + self.c_hat) + self.gamma_osc * (1.0 - self.c_hat)
        )

        self.contraction_rate = self.eta + np.abs(self.current_d) * (
            self.gamma - self.eta
        )

        self.r = self.contraction_rate * self.r

    def _prune(self, new_bounds: np.ndarray, global_bounds: np.ndarray) -> np.ndarray:
        Z = np.copy(new_bounds)
        Z[:, 1] = global_bounds[:, 0]
        Z.sort(axis=1)
        new_bounds[:, 0] = Z[:, 1]

        Z = np.copy(new_bounds)
        Z[:, 0] = global_bounds[:, 1]
        Z.sort(axis=1)
        new_bounds[:, 1] = Z[:, 0]
        new_bounds.sort(axis=1)

        for i, entry in enumerate(new_bounds):
            window_width = abs(entry[0] - entry[1])
            if window_width < self.min_window[i]:
                new_bounds[i, 0] -= (self.min_window[i] - window_width) / 2.0
                new_bounds[i, 1] += (self.min_window[i] - window_width) / 2.0

        return new_bounds

    def update(self, event, subject, *args, **kwargs):
        """Update reducer as subscriber to oracle."""

        if self.bounds.size == 0:
            self.init(subject.search_space)

        self.refit(subject.X[subject.y.argmin()])

        new_bounds = np.array(
            [
                self.current_optimal - 0.5 * self.r,
                self.current_optimal + 0.5 * self.r,
            ]
        ).T

        self.notify("on_suggest_search_space", new_bounds)
        self._prune(new_bounds, self.original_bounds)

        assert all(
            np.less(new_bounds[:, 0], new_bounds[:, 1])
        ), f"Incorrect bounds: {new_bounds}"

        self.bounds = np.concatenate([self.bounds, new_bounds])
        subject.bounds = new_bounds



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging

import click
from rich.logging import RichHandler
from rich.traceback import install


def init(verbose=False):
    """
    Init logging.

        Args:
            verbose (bool):
                Verbose level (DEBUG) or not (INFO).

        Raises:
            AnyError: If anything bad happens.

    """

    install(suppress=[click])

    logging.basicConfig(
        level="DEBUG" if verbose else "INFO",
        format="%(message)s",
        datefmt="[%X]",
        handlers=[
            RichHandler(
                rich_tracebacks=True,
                markup=True,
                enable_link_path=False,
            )
        ],
    )

    excluded_loggers = (
        "numba",
        "matplotlib",
        "executor",
        "sqlalchemy",
    )

    for log_name in excluded_loggers:
        other_log = logging.getLogger(log_name)
        other_log.setLevel(logging.WARNING)



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
from functools import partial
from typing import List, Optional

from feijoa.models.configuration import Configuration
from feijoa.search.oracles.oracle import Oracle

log = logging.getLogger(__name__)


class SeedOracle(Oracle):
    """
    Seed oracle used for measuring.

    Some known configurations and give a
    good basis for the rest of the oracles.

    Args:
        seeds:
            List of seed configurations.

    Raises:
        AnyError: If anything bad happens.

    """

    anchor = "seed"
    aliases = (
        "SeedOracle",
        "seed",
    )

    def __init__(self, *seeds, **kwargs):
        super().__init__(*seeds, **kwargs)
        self.seeds: list = list(seeds)
        self.is_emitted = False

    def ask(self, n: int = 1) -> Optional[List[Configuration]]:
        log.debug(
            "Parameter `n` does not affect on configuration's count,"
            f" because {self.__class__.__name__} is sequential."
        )

        cf = partial(Configuration, requestor=self.name)

        if not self.is_emitted:
            self.is_emitted = True
            return [cf(s) for s in self.seeds]
        else:
            return None

    def tell(self, config, result):
        # Tell no needed
        pass
